#include "sd_cmd.h"
#include "sd_spi.h"

#include <stdint.h>
#include <string.h>

// interne Variablen Definition ***********************************************************************

#define SD_BYTE_IDLE 0xFF

/** \brief untersttzte SD Kommandos dieser Lib
*/
enum sd_commands {
	SD_CMD00__GO_IDLE_STATE = 0,
	SD_CMD08__SEND_IF_COND = 8,
	SD_CMD09__SEND_CSD = 9,
	SD_CMD12__STOP_TRANSMISSION = 12,
	SD_CMD16__SET_BLOCKLEN = 16,
	SD_CMD17__READ_SINGLE_BLOCK = 17,
	SD_CMD18__READ_MULTIPLE_BLOCK = 18,
	SD_CMD24__WRITE_BLOCK = 24,
	SD_CMD25__WRITE_MULTIPLE_BLOCK = 25,
	SD_ACMD41__APP_SEND_OP_COND = 41,
	SD_CMD55__APP_CMD = 55,
	SD_CMD58__READ_OCR = 58,
};

enum sd_responses {
	SD_RES_R1,
	SD_RES_R1B,
	SD_RES_R3,
	SD_RES_R7,
};

enum sd_response_errs {
	SD_RES_NO_ERR = 0x00,
	SD_RES_IDLESTATE = 0x01,
	SD_RES_ILLEGALCOMMAND = 0x04,
};

/** \brief struct beinhaltet die Antwort eines SD Kommandos
*/
struct sd_cmd_response {
	enum sd_responses restype;
	union {
		struct {
			enum sd_response_errs val;
		} r1;
		struct {
			enum sd_response_errs val;
		} r1b;
		struct {
			enum sd_response_errs val;
			uint32_t ocr;
		} r3;
		struct {
			enum sd_response_errs val;
			uint32_t ocr;
		} r7;
	} res;
};

/** \brief Antworten der SD Karte auf ein Write Datapacket
*/
enum sd_data_responses {
	SD_DATARES_ACCEPTED = 0x05,
	SD_DATARES_ERR_CRC = 0x0B,
	SD_DATARES_ERR_WRITE = 0x0D,
};

/** \brief Header eines Datenpaketes
*/
enum sd_data_tokens {
	SD_DATA_TOKEN__READ_SINGLE_BLOCK = 0xFE,
	SD_DATA_TOKEN__READ_MULTIPLE_BLOCK = 0xFE,
	SD_DATA_TOKEN__WRITE_SINGLE_BLOCK = 0xFE,
	SD_DATA_TOKEN__WRITE_MULTIPLE_BLOCK = 0xFC,
	SD_DATA_TOKEN__WRITE_MULTIPLE_BLOCK_STOP = 0xFD,
};

// interne Variablen **********************************************************************************

static struct sd_cardinfo cardinfo;

// Private Funktionen Definition **********************************************************************

/** \brief initialisiert eine als SDv1 oder MMC Karte erkannte Karte
	\return SD_OK oder SD_ERR
*/
static
enum sd_errcodes sd_init_sdmmc(void);

/** \brief initialisiert eine als SDv2 erkannte Karte
	\return SD_OK oder SD_ERR
*/
static
enum sd_errcodes sd_init_sdv2(void);

/** \brief empfngt ein Data Packet von der SD Karte
	\param buf speicher fr empfangende Daten
	\param token welches Data Packet solls sein? -> SINGLE/MULTIPLE
	\param blocksize Anzahl der zu sendenden Bytes
*/
static
enum sd_errcodes sd_write_datapacket(const uint8_t *buf, enum sd_data_tokens token, uint32_t blocksize);

/** \brief sendet ein Data Packet zur SD Karte und wertet die Data Response aus
	\param buf zu sendende Daten
	\param token welches Data Packet solls sein? -> SINGLE/MULTIPLE
	\param blocksize Anzahl der zu empfangenen Bytes
*/
static
enum sd_errcodes sd_read_datapacket(uint8_t *buf, enum sd_data_tokens token, uint32_t blocksize);

/** \brief setzt CS HIGH und sendet das Idle Byte zum SD Do loslassen
	\param err auszugebender Error oder SD_OK
	\return err
*/
static
enum sd_errcodes sd_errout(enum sd_errcodes err);

/** \brief wartet solange bis das Byte auf dem SPI Bus auftaucht
	\param byte auf dieses Byte (bzw. Bitmuster) wird gewartet
*/
static
enum sd_errcodes sd_waitfor(uint8_t byte);

/** \brief wartet auf x IDLE Bytes bis die SD Command Antwort kommt
*/
static
uint8_t sd_waitfor_answer(void);

/** \brief sendet ein Kommando an die SD Karte
	\param command das auszufhrende Kommando
	\param arg das Argument zum Kommando oder 0 wenn nicht bentigt
	\param ans Pointer zum Antwortstruct
*/
static
void sd_command(enum sd_commands command, uint32_t arg, struct sd_cmd_response *ans);

/** \brief liet das CSD Register und bergibt es dem v1 oder v2 parser
	\return SD_OK oder SD_ERR
*/
static 
enum sd_errcodes sd_get_csd(void);

/** \brief CSDv1 parsen
	\param csd BUffer der 16 Bytes des CSD Registers
*/
static
enum sd_errcodes sd_parse_csdv1(uint8_t *csd);

/** \brief CSDv2 parsen
	\param csd BUffer der 16 Bytes des CSD Registers
*/
static
enum sd_errcodes sd_parse_csdv2(uint8_t *csd);

/** \brief extrahiert das Bitfield aus einem Buffer
	\param buf Buffer des Bitfields
	\param bit_oben Bit bis zu dem gelesen werden soll
	\param bit_unten Bit ab dem gelesen werden soll
	\return den Inhalt des ausgewhlten Bitfields
*/
static
uint32_t sd_get_bitfield(uint8_t *buf, uint8_t bit_oben, uint8_t bit_unten);

// ffentliche Funktionen *****************************************************************************

enum sd_errcodes sd_init_card(void){

	struct sd_cmd_response response;
	
	// SPI Init zwischen 100kHz und 400kHz
	sd_spi_init(SPI_CLKDIV_LOW);
	
	// 74+ Taktzyklen lang braucht die SD Karte fr ihren Init -> 10 Dummybytes senden 
	unsigned int i;
	for (i = 0; i < 10; i++){
		sd_spi_iobyte(SD_BYTE_IDLE);
	}
	
	memset(&cardinfo, 0, sizeof(cardinfo));
	
	sd_spi_setcs(0);
	
	// als erstes die Karte in den Idle State versetzen
	sd_command(SD_CMD00__GO_IDLE_STATE, 0, &response);
	if (response.res.r1.val != SD_RES_IDLESTATE){
		return sd_errout(SD_ERR);
	}
	
	// Jetzt mit CMD8 die Brechstange ansetzen, hierrauf Antworten nur SDv2 Karten sinnvoll
	sd_command(SD_CMD08__SEND_IF_COND, 0x1AA, &response);
	if (response.res.r7.val == (SD_RES_ILLEGALCOMMAND | SD_RES_IDLESTATE)){
		// Diese Karte kennt CMD8 nicht -> SDv1 oder MMC
		return sd_init_sdmmc();
	}
	if ((response.res.r7.val == SD_RES_IDLESTATE) &&
		(response.res.r7.ocr & 0xFFF) == 0x1AA){
		return sd_init_sdv2();
	}
	
	// wenn hier angekommen lief was schief
	return sd_errout(SD_ERR);
}

struct sd_cardinfo* sd_get_cardinfo(void){
	return &cardinfo;
}

enum sd_errcodes 
sd_read_blocks(uint8_t *buf, uint32_t lba, uint32_t block_count, uint32_t blocksize){

	struct sd_cmd_response response;
	enum sd_errcodes data_err;
  
	// SDHC nutzt Blockadressen, SD/MMC nutzt Byteadressen
	unsigned int sdcard_addr = lba;
	if (cardinfo.cardtype != SD_CARDTYPE_SDHC){
		sdcard_addr *= blocksize;
	}
	
	sd_spi_setcs(0);
	
	if (block_count == 1){
		
		// CMD17 senden um einen Block zu lesen, nur weitermachen wenn die SD Karte ihr OK gibt
		sd_command(SD_CMD17__READ_SINGLE_BLOCK, sdcard_addr, &response);
		if (response.res.r1.val != SD_RES_NO_ERR){
			return sd_errout(SD_ERR);
		}
		
		data_err = sd_read_datapacket(buf, SD_DATA_TOKEN__READ_SINGLE_BLOCK, blocksize);
		if (data_err != SD_OK){
			return sd_errout(SD_ERR);
		}
		
	}else{ // block_count
		
		// CMD18 senden um mehrere Blcke zu lesen, nur weitermachen wenn die SD Karte ihr OK gibt
		sd_command(SD_CMD18__READ_MULTIPLE_BLOCK, sdcard_addr, &response);
		if (response.res.r1.val != SD_RES_NO_ERR){
			return sd_errout(SD_ERR);
		}
		  
		while (block_count--){
		
			data_err = sd_read_datapacket(buf, SD_DATA_TOKEN__READ_MULTIPLE_BLOCK, blocksize);
			if (data_err != SD_OK){
				return sd_errout(SD_ERR);
			}
				
			// neuen Datenpointer berechnen zum Daten einlesen
			buf += blocksize;
		}
		
		// warten auf nchsten Block, die SD Karte sucht intern den neuen Block raus
		// und kann daher nicht auf den STOP Command reagieren
		if (sd_waitfor(SD_DATA_TOKEN__READ_MULTIPLE_BLOCK) != SD_OK){
			return sd_errout(SD_ERR);
		}
		
		// CMD12, wir wollen keine weiteren Blcke mehr lesen
		sd_command(SD_CMD12__STOP_TRANSMISSION, 0, &response);
		if (response.res.r1b.val != SD_RES_NO_ERR){
			return sd_errout(SD_ERR);
		}
		
		// warten bis SD DO wieder 0xFF ist, die Karte ist busy nach CMD12
		while (sd_spi_iobyte(SD_BYTE_IDLE) != SD_BYTE_IDLE);
	}
	
	sd_spi_setcs(1);
	sd_spi_iobyte(SD_BYTE_IDLE);
  
	return SD_OK;
}

enum sd_errcodes
sd_write_blocks(const uint8_t *buf, uint32_t lba, uint32_t block_count, uint32_t blocksize){
	
	struct sd_cmd_response response;
	enum sd_errcodes dataresponse_err;
	
	// SDHC nutzt Blockadressen, SD/MMC nutzt Byteadressen
	unsigned int sdcard_addr = lba;
	if (cardinfo.cardtype != SD_CARDTYPE_SDHC){
		sdcard_addr *= blocksize;
	}
	
	sd_spi_setcs(0);
	
	if (block_count == 1){
		
		// CMD24 senden um einen Block zu schreiben nur weitermachen wenn die SD Karte ihr OK gibt
		sd_command(SD_CMD24__WRITE_BLOCK, sdcard_addr, &response);
		if (response.res.r1.val != SD_RES_NO_ERR){
			return sd_errout(SD_ERR);
		}
		
		// mind 1 Dummybyte senden bevor der Datenblock kommt, aus Sicherheitsgrnden sinds 2
		sd_spi_iobyte(SD_BYTE_IDLE);
		sd_spi_iobyte(SD_BYTE_IDLE);
		
		dataresponse_err = sd_write_datapacket(buf, SD_DATA_TOKEN__WRITE_SINGLE_BLOCK, blocksize);
		if (dataresponse_err != SD_OK){
			return sd_errout(dataresponse_err);
		}
		
	}else{ // block_count
		
		// CMD25 senden um mehrere Blcke zu schreiben, nur weitermachen wenn die SD Karte ihr OK gibt
		sd_command(SD_CMD25__WRITE_MULTIPLE_BLOCK, sdcard_addr, &response);
		if (response.res.r1.val != SD_RES_NO_ERR){
			return sd_errout(SD_ERR);
		}
		
		// mind 1 Dummybyte senden bevor der Datenblock kommt, aus Sicherheitsgrnden sinds 2
		sd_spi_iobyte(SD_BYTE_IDLE);
		sd_spi_iobyte(SD_BYTE_IDLE);
		
		while (block_count--){
			
			dataresponse_err = sd_write_datapacket(buf, SD_DATA_TOKEN__WRITE_MULTIPLE_BLOCK, blocksize);
			if (dataresponse_err != SD_OK){
				return sd_errout(dataresponse_err);
			}
			
			// neuen Datenpointer berechnen zum Daten einlesen
			buf += blocksize;
			
		}
		
		// Data Stop Token senden
		sd_spi_iobyte(SD_DATA_TOKEN__WRITE_MULTIPLE_BLOCK_STOP);
		
		// 1 Byte warten
		sd_spi_iobyte(SD_BYTE_IDLE);
		
		// warten bis Busy Flag gelscht 
		while (sd_spi_iobyte(SD_BYTE_IDLE) != SD_BYTE_IDLE);
	}


	// 1 Byte Wartezyklen senden damit die Karte MISO loslsst
	sd_spi_setcs(1);    
	sd_spi_iobyte(SD_BYTE_IDLE);

	return SD_OK;
}

// Private Funktionen **********************************************************************************

static
enum sd_errcodes sd_init_sdmmc(void){
	
	struct sd_cmd_response response;
	
	do{
		
		// ACMD41 senden, SDv1 antwortet mit Idle State, aber MMC mit Illegal Command
		// vor einem ACMD muss CMD55 gesendet werden 
		sd_command(SD_CMD55__APP_CMD, 0, &response);
		sd_command(SD_ACMD41__APP_SEND_OP_COND, 0, &response);
		if (response.res.r1.val & SD_RES_ILLEGALCOMMAND){
			// das hier ist eine NICHT untersttzte MMC Karte
			cardinfo.cardtype = SD_CARDTYPE_MMC;
			return sd_errout(SD_ERR);
		}
		
	}while(response.res.r1.val == SD_RES_IDLESTATE);
	
	// hier angekommen kann es nur eine SDv1 Karte sein
	cardinfo.cardtype = SD_CARDTYPE_SDV1;
	
	// bei 2GB SDv1 Karten kann die Blocksize wohl mal auf 1024 Byte gesetzt sein
	// aber elm chan FATfs will 512Byte
	sd_command(SD_CMD16__SET_BLOCKLEN, 512, &response);
	if (response.res.r1.val != SD_RES_NO_ERR){
		return sd_errout(SD_ERR);
	}
	
	return sd_get_csd();
}

static
enum sd_errcodes sd_init_sdv2(void){
	
	struct sd_cmd_response response;
	
	// Initialisierungsschleife
	do{
		
		// vor einem ACMD muss CMD55 gesendet werden 
		sd_command(SD_CMD55__APP_CMD, 0, &response);
		
		// ACMD41 senden fr den Init und checken ob SDHC Karte (HCS Flag)
		sd_command(SD_ACMD41__APP_SEND_OP_COND, (1 << 30), &response);
		
	}while(response.res.r1.val == SD_RES_IDLESTATE);
	
	sd_command(SD_CMD58__READ_OCR, 0, &response);
	if (response.res.r3.val != SD_RES_NO_ERR){
		return sd_errout(SD_ERR);
	}
	
	cardinfo.cardtype = SD_CARDTYPE_SDv2;
	
	// ist das hier ne SDHC Karte mit Blockadressierung? (SDHC hat immer 512 Byte Blcke!)
	if(response.res.r3.ocr & (1 << 30)){
		cardinfo.cardtype = SD_CARDTYPE_SDHC;
		cardinfo.blocksize = 512;
	}
	
	return sd_get_csd();
}

static
enum sd_errcodes sd_write_datapacket(const uint8_t *buf, enum sd_data_tokens token, uint32_t blocksize){
	
	// Data Start Token senden
	sd_spi_iobyte(token);
	
	// Daten des Blocks schreiben
	sd_spi_write(buf, blocksize);
	
	// CRC wurde abgeschalten, aber Dummys mssen den Frame auffllen
	sd_spi_iobyte(SD_BYTE_IDLE);
	sd_spi_iobyte(SD_BYTE_IDLE);
	
	// Data Response lesen nach den gesendeten Daten
	enum sd_data_responses response;
	response = (enum sd_data_responses)(0x1F & sd_spi_iobyte(SD_BYTE_IDLE));
	
	switch (response){
		
		case SD_DATARES_ERR_CRC:
			return SD_ERR_CRC;
			
		case SD_DATARES_ERR_WRITE:
			return SD_ERR_WRITE;
			
		case SD_DATARES_ACCEPTED:
			// warten bis Busy Flag gelscht
			while (sd_spi_iobyte(SD_BYTE_IDLE) != SD_BYTE_IDLE);
			return SD_OK;
	}
	
	return SD_ERR;
}

static
enum sd_errcodes sd_read_datapacket(uint8_t *buf, enum sd_data_tokens token, uint32_t blocksize){

	// auf das Data Start Token warten
	if (sd_waitfor(token) != SD_OK){
		return SD_ERR;
	}
	
	// Daten des Blocks einlesen
	sd_spi_read(buf, blocksize);
		
	// CRC wegwerfen
	sd_spi_iobyte(SD_BYTE_IDLE);
	sd_spi_iobyte(SD_BYTE_IDLE);
	
	return SD_OK;
}

static
enum sd_errcodes sd_errout(enum sd_errcodes err){

	// 1 Byte Wartezyklen senden damit die Karte MISO loslsst
	sd_spi_setcs(1);
	sd_spi_iobyte(SD_BYTE_IDLE);
	
	return err;
}

static
enum sd_errcodes sd_waitfor(uint8_t byte){

	uint32_t maxtry = 66666;
	uint32_t tries = 0;
	uint8_t readbyte;
	
	while(tries < maxtry){
		readbyte = sd_spi_iobyte(SD_BYTE_IDLE);
		if (readbyte == byte){
			break;
		}
		tries++;
	}
	
	if (tries == maxtry){
		return SD_ERR;
	}

	return SD_OK;
}

static
uint8_t sd_waitfor_answer(void){
  
	uint32_t maxtry = 66;
	uint32_t tries = 0;
	uint8_t readbyte;
 
	while(tries < maxtry){
		readbyte = sd_spi_iobyte(SD_BYTE_IDLE);
		if (readbyte != SD_BYTE_IDLE){
			break;
		}
		tries++;
	}

	return readbyte;
}

static
void sd_command(enum sd_commands command, uint32_t arg, struct sd_cmd_response *ans){
	
	// vorgemerkte CRC Summen fr Init Befehle oder faken
	uint8_t crc = 0xFF;
	if (command == SD_CMD00__GO_IDLE_STATE){
		crc = 0x95;
	}
	if (command == SD_CMD08__SEND_IF_COND){
		crc = 0x87;
	}
	
	// Kommando zusammenbauen
	uint8_t commandbuf[6];
	commandbuf[0]  = 0b01000000; 			//Bit 7 immer 0, Bit 6 immer gesetzt
	commandbuf[0] |= (uint8_t)command;
	commandbuf[1]  = (uint8_t)(arg >> 24);	//Argument aufteilen
	commandbuf[2]  = (uint8_t)(arg >> 16);
	commandbuf[3]  = (uint8_t)(arg >> 8);
	commandbuf[4]  = (uint8_t)(arg);
	commandbuf[5]  = (crc | 0x01);			//CRC (Bit 0 immer gesetzt)
	
	// Kommando senden
	sd_spi_write(commandbuf, 6);
	
	// je nach Kommando die passende Antwort parsen
	switch (command){
		
		// diese Kommandos antworten R1
		case SD_CMD00__GO_IDLE_STATE:
		case SD_CMD09__SEND_CSD:
		case SD_CMD16__SET_BLOCKLEN:
		case SD_CMD17__READ_SINGLE_BLOCK:
		case SD_CMD18__READ_MULTIPLE_BLOCK:
		case SD_CMD24__WRITE_BLOCK:
		case SD_CMD25__WRITE_MULTIPLE_BLOCK:
		case SD_ACMD41__APP_SEND_OP_COND:
		case SD_CMD55__APP_CMD:
			ans->restype = SD_RES_R1;
			ans->res.r1.val = sd_waitfor_answer();
			break;
			
		// diese Kommandos antworten R1b
		case SD_CMD12__STOP_TRANSMISSION:
			//zuerst kommt 1 Byte Mll (alter Datenblock wird noch gesendet)
			sd_spi_iobyte(SD_BYTE_IDLE);
			//dann die Antwort mit eventuell nach etwas Bedenkzeit
			ans->restype = SD_RES_R1B;
			ans->res.r1b.val = sd_waitfor_answer();
			break;
			
		case SD_CMD58__READ_OCR:
			ans->restype = SD_RES_R3;
			ans->res.r3.val = sd_waitfor_answer();
			ans->res.r3.ocr = 0;
			ans->res.r3.ocr |= sd_spi_iobyte(SD_BYTE_IDLE) << 24;
			ans->res.r3.ocr |= sd_spi_iobyte(SD_BYTE_IDLE) << 16;
			ans->res.r3.ocr |= sd_spi_iobyte(SD_BYTE_IDLE) << 8;
			ans->res.r3.ocr |= sd_spi_iobyte(SD_BYTE_IDLE);
			break;
			
		case SD_CMD08__SEND_IF_COND:
			ans->restype = SD_RES_R7;
			ans->res.r7.val = sd_waitfor_answer();
			ans->res.r7.ocr = 0;
			ans->res.r7.ocr |= sd_spi_iobyte(SD_BYTE_IDLE) << 24;
			ans->res.r7.ocr |= sd_spi_iobyte(SD_BYTE_IDLE) << 16;
			ans->res.r7.ocr |= sd_spi_iobyte(SD_BYTE_IDLE) << 8;
			ans->res.r7.ocr |= sd_spi_iobyte(SD_BYTE_IDLE);
			break;
	}
}

static 
enum sd_errcodes sd_get_csd(void){
	
	struct sd_cmd_response response;
	uint8_t csd[16];
	int8_t i;
	
	// Takt auf Vollgas schalten
	sd_spi_setcs(1);
	sd_spi_iobyte(SD_BYTE_IDLE);
	
	sd_spi_init(SPI_CLKDIV_HIGH);
	
	sd_spi_setcs(0);
	
	sd_command(SD_CMD09__SEND_CSD, 0, &response);
	if (response.res.r1.val != SD_RES_NO_ERR){
		return sd_errout(SD_ERR);
	}
	
	// auf das Data Start Token warten
	if (sd_waitfor(SD_DATA_TOKEN__READ_SINGLE_BLOCK) != SD_OK){
		return sd_errout(SD_ERR);
	}
	
	// jetzt die 16 Byte des CSD einlesen, oberste Bytes kommen zuerst
	for(i = 15; i > 0; i--){
		csd[i] = sd_spi_iobyte(SD_BYTE_IDLE);
	}
	
	// CRC wegwerfen
	sd_spi_iobyte(SD_BYTE_IDLE);
	sd_spi_iobyte(SD_BYTE_IDLE);
	
	// 1 Byte Wartezyklen senden damit die Karte MISO loslsst
	sd_spi_setcs(1);
	sd_spi_iobyte(SD_BYTE_IDLE);
	
	uint8_t CSD_STRUCTURE = sd_get_bitfield(csd, 127, 126);
	
	if (CSD_STRUCTURE == 0){
		return sd_parse_csdv1(csd);
	}
	
	if (CSD_STRUCTURE == 1){
		return sd_parse_csdv2(csd);
	}
	
	// CSD konnte nicht geparsed werden
	return SD_ERR;
}

static
enum sd_errcodes sd_parse_csdv1(uint8_t *csd){

	cardinfo.READ_BL_LEN = sd_get_bitfield(csd, 83, 80);
	uint32_t C_SIZE = sd_get_bitfield(csd, 73, 62);
	uint32_t C_SIZE_MULT = sd_get_bitfield(csd, 49, 47);

	cardinfo.blocksize = 0;
	if (cardinfo.READ_BL_LEN == 9){
		cardinfo.blocksize = 512;
	}
	if(cardinfo.READ_BL_LEN == 10){
		cardinfo.blocksize = 1024;
	}
	if(cardinfo.READ_BL_LEN == 11){
		cardinfo.blocksize = 2048;
	}
	if (cardinfo.blocksize == 0){
		return SD_ERR;
	}

	uint32_t i, MULT = 1;
	for(i = 0; i < C_SIZE_MULT+2; i++){
		MULT *= 2;
	}

	cardinfo.block_nbr = (C_SIZE+1) * MULT;
	cardinfo.memory_capacity = cardinfo.block_nbr * cardinfo.blocksize;

	return SD_OK;
} 

static
enum sd_errcodes sd_parse_csdv2(uint8_t *csd){
	
	// Blockgre als 2^READ_BL_LEN
	cardinfo.READ_BL_LEN = sd_get_bitfield(csd, 83, 80);
	
	// Speichergre der Karte
	cardinfo.memory_capacity = sd_get_bitfield(csd, 69, 48);
	cardinfo.memory_capacity = (cardinfo.memory_capacity+1) * (1024*512);
	cardinfo.block_nbr = cardinfo.memory_capacity/512;
	
	return SD_OK;
}

static
uint32_t sd_get_bitfield(uint8_t *buf, uint8_t bit_oben, uint8_t bit_unten){

	uint8_t fieldlen = bit_oben - bit_unten + 1;
	if (fieldlen > 32){
		return 0;
	} 
	
	uint32_t i, ret = 0;
	uint8_t bufindex = bit_unten/8; 				// welcher Index im Buffer
	uint8_t bitindex = bit_unten - (bufindex*8); 	// welches Bit im jetzigen Bufferindex
	
	for (i = 0; i < fieldlen; i++){
		if (buf[bufindex] & (1 << bitindex)){
			ret |= (1 << i);
		}
		//deprintf("bitparse: i=%u, bufi=%u, biti=%u\n", i, bufindex, bitindex);
		bitindex++;
		if (bitindex == 8){
			bitindex = 0;
			bufindex++;
		}
	}
	
	return ret;
}