#include "spi.h"
#include "spi_regdefs.h"

#if SPI_USE_DMA
	#include "armcore/nvic.h"
	#include "periph/dma.h"
#endif

static const unsigned int spi_bases[3] =
	{SPI1_BASE_INT, SPI2_BASE_INT, SPI3_BASE_INT}; 

#if SPI_USE_DMA	
static const uint8_t spi_irq_nbrs[3] =
	{IRQ_NBR_SPI1, IRQ_NBR_SPI2, IRQ_NBR_SPI3}; 
	
static volatile void *callb_irq_data[3] = {NULL, NULL, NULL};
static void (*callb_irq[3])(unsigned int, void*)  = {NULL, NULL, NULL};
#endif

//*************************************** SPI Abteilung **********************************************

void spi_init_master(
	unsigned int spi_base, 	//SPI Base
	unsigned int lsbfirst, 	//LSB first = 1
	enum spi_clkdiv div,	//CLK div
	unsigned int cpol,		//CPOL
	unsigned int cpha)		//CPHA
	{
	
	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	spi->I2SCFGR = 0; //falls vorher I2S aktiv war
	unsigned int cr1 = 0;
	
	if (lsbfirst){
		cr1 |= CR1_LSBFIRST;
	}
	
	cr1 |= div << CR1_BR_SHIFT;
	
	if (cpol){
		cr1 |= CR1_CPOL;
	}
	
	if (cpha){
		cr1 |= CR1_CPHA;
	}
	
	spi->CR1 = cr1 | CR1_MSTR;
	spi->CR2 = CR2_SSOE;
	spi->CR1 |= CR1_SPE;
}

uint8_t spi_rw_byte(unsigned int spi_base, uint8_t byte){

	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	
	while(0 == (spi->SR & SR_TXE));
	spi->DR = byte;
	while(0 == (spi->SR & SR_RXNE));
	return spi->DR;
}

void spi_rw_bytes(unsigned int spi_base, unsigned int bytes, uint8_t *tx, uint8_t *rx){
	
	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	unsigned int i;
	
	for (i = 0; i < bytes; i++){
		while(0 == (spi->SR & SR_TXE));
		spi->DR = tx[i];
		while(0 == (spi->SR & SR_RXNE));
		rx[i] = spi->DR;
	}
}

void spi_w_bytes(unsigned int spi_base, unsigned int bytes, uint8_t *tx){

	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	unsigned int i;
	unsigned int devnull;

	for (i = 0; i < bytes; i++){
		while(0 == (spi->SR & SR_TXE));
		spi->DR = tx[i];
		while(0 == (spi->SR & SR_RXNE));
		devnull = spi->DR;
	}
	(void)devnull;
}

void spi_r_bytes(unsigned int spi_base, unsigned int bytes, uint8_t idle_byte, uint8_t *rx){
	
	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	unsigned int i;
	
	for (i = 0; i < bytes; i++){
		while(0 == (spi->SR & SR_TXE));
		spi->DR = idle_byte;
		while(0 == (spi->SR & SR_RXNE));
		rx[i] = spi->DR;
	}
}

//Callback reinhngen fr "DMA hat alle Kanle durch"
#if SPI_USE_DMA
int spi_dmamode(
	unsigned int spi_base, 							//Welcher SPI
	uint8_t *tx, 									//SPI Sendebuffer
	uint8_t *rx,									//SPI Empfangsbuffer
	enum dma_periph stream_tx,						//DMA Sendestream
	enum dma_periph stream_rx,						//DMA Empfangsstream
	uint16_t buflen,								//Anzahl der zu bertragenden Daten
	void (*callb_dma)(unsigned int, void*), 		//IRQ Handler DMA
	void *callb_data_dma, 							//IRQ Handler DMA Daten
	void (*callb)(unsigned int, void*), 			//IRQ Handler SPI Err
	void *callb_data){								//IRQ Handler SPI Err Daten

	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	
	spi->CR2 &= ~(CR2_TXDMAEN | CR2_RXDMAEN);
	asm("DSB");

	int dma_arr;
	
	//RX DMA zuerst einstellen, falls DMA Bits noch gesetzt sind
	dma_arr = dma_periph(
		stream_rx,			//Angabe der Peripherie
		(unsigned int)rx,
		(unsigned int)&(spi->DR),
		buflen,				//Anzahl der zu bertragenden Elemente
		DMA_SIZE_BYTE,		//Gre der zu bertragenden Daten
		0,					//nicht circular
		DMA_PERIPH_TO_MEM,	//in welche Richtung geht der DMA (mem to mem ist verboten)
		DMA_PRIO_HIGH,		//Prioritt des DMA
		callb_dma,			//Callback bei auftretendem IRQ
		callb_data_dma		//Daten fr den Callback
		);
	if (dma_arr){
		return -1;
	}
	
	//TX DMA
	dma_arr = dma_periph(
		stream_tx,			//Angabe der Peripherie
		(unsigned int)tx,
		(unsigned int)&(spi->DR),
		buflen,				//Anzahl der zu bertragenden Elemente
		DMA_SIZE_BYTE,		//Gre der zu bertragenden Daten
		0,					//nicht circular
		DMA_MEM_TO_PERIPH,	//in welche Richtung geht der DMA (mem to mem ist verboten)
		DMA_PRIO_LOW,		//Prioritt des DMA
		callb_dma,			//Callback bei auftretendem IRQ
		callb_data_dma		//Daten fr den Callback
		);
	if (dma_arr){
		return -1;
	}
	
	callb_irq[spi_base] = callb;
	callb_irq_data[spi_base] = callb_data;
	
	spi->CR2 |= CR2_ERRIE | CR2_TXDMAEN | CR2_RXDMAEN;
	
	//IRQ aktivieren
	nvic_enable_irq(spi_irq_nbrs[spi_base], 0xF, 0xF);
	
	return 0;
}
#endif

//*************************************** I2S Abteilung **********************************************

void spi_i2s_init(
	unsigned int spi_base, 
	enum SPI_I2S_MODE i2smode, 
	enum SPI_I2S_STD standard, 
	enum SPI_I2S_DATLEN datlen,
	unsigned int mcko_enable,
	enum dma_periph peri,		//Angabe der Peripherie
	unsigned int mem_addr0,
	unsigned int mem_addr1,
	uint16_t data_items_nbr,	//Anzahl der zu bertragenden Elemente
	void (*callb)(unsigned int, void*),	//Callback bei auftretendem IRQ
	void *callb_data						//Daten fr den Callback
	){

	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	
	//SPI abschalten
	spi->CR1 = 0;
	spi->CR2 = 0;
	
	unsigned int SPI_I2SPR = 0;
	
	if (mcko_enable){
		SPI_I2SPR |= I2SPR_MCKOE;
	}
	SPI_I2SPR |= 2 << I2SPR_I2SDIV_SHIFT; // I2s_CLK/4 = MCK
	spi->I2SPR = SPI_I2SPR;
	
	unsigned int SPI_I2SCFGR = 0;
	SPI_I2SCFGR |= I2SCFGR_I2SMOD; // I2S Modus aktivieren
	SPI_I2SCFGR |= i2smode << I2SCFGR_I2SCFG_SHIFT;
	SPI_I2SCFGR |= standard << I2SCFGR_I2SSTD_SHIFT;
	SPI_I2SCFGR |= datlen << I2SCFGR_DATLEN_SHIFT;
	SPI_I2SCFGR |= I2SCFGR_CHLEN; // immer 32Bit wegen Takteinhaltung
	spi->I2SCFGR = SPI_I2SCFGR;
	
	//DMA init
	int dma_res = dma_periph_dbm(
		peri,				//Angabe der Peripherie
		mem_addr0,
		mem_addr1,
		(unsigned int)&spi->DR,
		data_items_nbr,		//Anzahl der zu bertragenden Elemente
		DMA_SIZE_HWORD,		//Gre der zu bertragenden Daten
		DMA_MODE_DBUF,		//mem wird circulr befllt = DMA geht nie zuende
		DMA_MEM_TO_PERIPH,	//in welche Richtung geht der DMA (mem to mem ist verboten)
		DMA_PRIO_MED,		//Prioritt des DMA
		callb,				//Callback bei auftretendem IRQ
		callb_data			//Daten fr den Callback
		);
		
	if (dma_res){
		return;
	}
	
	spi->CR2 |= CR2_TXDMAEN;
	
	//alles einschalten
	spi->I2SCFGR |= I2SCFGR_I2SE;
	
	/*unsigned short out = 0; //0b1011110000001101;
	while(1){
		while(!(spi->SR & SR_TXE));
		spi->DR = out;
		while(!(spi->SR & SR_TXE));
		spi->DR = out;
		out += 20;
	}*/
}

void spi_i2s_write(unsigned int spi_base, unsigned short val){
	
	volatile struct spi * const spi = (struct spi *)spi_bases[spi_base];
	
	while(!(spi->SR & SR_TXE));
	spi->DR = val;
}

#if (SPI_USE_IRQ | SPI_USE_DMA)
static void irq_spi(unsigned int spi_nbr){
	
	nvic_pending_clear(spi_irq_nbrs[spi_nbr]);
	volatile struct spi * const spi = (struct spi *)spi_bases[spi_nbr];
	
	if (callb_irq[spi_nbr]){
		callb_irq[spi_nbr](spi->SR, (void*)callb_irq_data[spi_nbr]);
	}
}

void irq_spi1(void){
	irq_spi(0);
}

void irq_spi2(void){
	irq_spi(1);
}

void irq_spi3(void){
	irq_spi(2);
}
#endif
